/* table.c
**
**    create the billard-table display lists
**    Copyright (C) 2001  Florian Berger
**    Email:  harpin_floh@yahoo.de,  florian.berger@jk.uni-linz.ac.at
**
**    This program is free software; you can redistribute it and/or modify
**    it under the terms of the GNU General Public License Version 2 as
**    published by the Free Software Foundation;
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program; if not, write to the Free Software
**    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
**
*/

#include <GL/glu.h>
#include <GL/gl.h>
#include <stdlib.h>
#include "billard.h"
#include "table.h"
#include "options.h"
#include "png_loader.h"


void my_glBox( float x1, float y1, float z1,
               float x2, float y2, float z2 )
{
    float dummy;

    if( x1 > x2 ){ dummy=x1; x1=x2; x2=dummy; }
    if( y1 > y2 ){ dummy=y1; y1=y2; y2=dummy; }
    if( z1 > z2 ){ dummy=z1; z1=z2; z2=dummy; }
    
    glBegin(GL_QUADS);
    
       glNormal3f( 0.0, -1.0, 0.0 );
       glVertex3f(x1,y1,z1); glVertex3f(x1,y1,z2);
       glVertex3f(x2,y1,z2); glVertex3f(x2,y1,z1);

       glNormal3f( 1.0, 0.0, 0.0 );
       glVertex3f(x2,y1,z1); glVertex3f(x2,y1,z2);
       glVertex3f(x2,y2,z2); glVertex3f(x2,y2,z1);
       
       glNormal3f( 0.0, 1.0, 0.0 );
       glVertex3f(x2,y2,z1); glVertex3f(x2,y2,z2);
       glVertex3f(x1,y2,z2); glVertex3f(x1,y2,z1);

       glNormal3f( -1.0, 0.0, 0.0 );
       glVertex3f(x1,y2,z1); glVertex3f(x1,y2,z2);
       glVertex3f(x1,y1,z2); glVertex3f(x1,y1,z1);

       glNormal3f( 0.0, 0.0, -1.0 );
       glVertex3f(x1,y1,z1); glVertex3f(x2,y1,z1);
       glVertex3f(x2,y2,z1); glVertex3f(x1,y2,z1);
       
       glNormal3f( 0.0, 0.0, 1.0 );
       glVertex3f(x1,y1,z2); glVertex3f(x1,y2,z2);
       glVertex3f(x2,y2,z2); glVertex3f(x2,y1,z2);
       
    glEnd();

}


void autonormalize_quad( VMvect v1, VMvect v2, VMvect v3, VMvect v4, int order )
{
    VMvect n;
    n=vec_unit(vec_cross(vec_diff(v2,v1),vec_diff(v3,v1)));
    glNormal3f( n.x, n.y, n.z );
    if( !order ) {
        glVertex3f(v4.x,v4.y,v4.z);
        glVertex3f(v3.x,v3.y,v3.z);
        glVertex3f(v2.x,v2.y,v2.z);
        glVertex3f(v1.x,v1.y,v1.z);
    } else {
        glVertex3f(v1.x,v1.y,v1.z);
        glVertex3f(v2.x,v2.y,v2.z);
        glVertex3f(v3.x,v3.y,v3.z);
        glVertex3f(v4.x,v4.y,v4.z);
    }
}


void autonormalize_triangle( VMvect v1, VMvect v2, VMvect v3, int order )
{
    VMvect n;
    n=vec_unit(vec_cross(vec_diff(v2,v1),vec_diff(v3,v1)));
    if( !order ) {
        glNormal3f( n.x, n.y, n.z );
        glVertex3f(v3.x,v3.y,v3.z);
        glNormal3f( n.x, n.y, n.z );
        glVertex3f(v2.x,v2.y,v2.z);
        glNormal3f( n.x, n.y, n.z );
        glVertex3f(v1.x,v1.y,v1.z);
    } else {
        glNormal3f( n.x, n.y, n.z );
        glVertex3f(v1.x,v1.y,v1.z);
        glNormal3f( n.x, n.y, n.z );
        glVertex3f(v2.x,v2.y,v2.z);
        glNormal3f( n.x, n.y, n.z );
        glVertex3f(v3.x,v3.y,v3.z);
    }
}


void autonormalize_triangle_round( VMvect v1, VMvect v2, VMvect v3, int order, double round )
{
    VMvect n0,c,n;
    n0=vec_cross(vec_diff(v2,v1),vec_diff(v3,v1));
    c=tri_center(v1,v2,v3);
    if( !order ) {
        n=vec_unit(vec_add(n0,vec_scale(vec_diff(v3,c),round))); glNormal3f( n.x, n.y, n.z );
        glVertex3f(v3.x,v3.y,v3.z);
        n=vec_unit(vec_add(n0,vec_scale(vec_diff(v2,c),round))); glNormal3f( n.x, n.y, n.z );
        glVertex3f(v2.x,v2.y,v2.z);
        n=vec_unit(vec_add(n0,vec_scale(vec_diff(v1,c),round))); glNormal3f( n.x, n.y, n.z );
        glVertex3f(v1.x,v1.y,v1.z);
    } else {
        n=vec_unit(vec_add(n0,vec_scale(vec_diff(v1,c),round))); glNormal3f( n.x, n.y, n.z );
        glVertex3f(v1.x,v1.y,v1.z);
        n=vec_unit(vec_add(n0,vec_scale(vec_diff(v2,c),round))); glNormal3f( n.x, n.y, n.z );
        glVertex3f(v2.x,v2.y,v2.z);
        n=vec_unit(vec_add(n0,vec_scale(vec_diff(v3,c),round))); glNormal3f( n.x, n.y, n.z );
        glVertex3f(v3.x,v3.y,v3.z);
    }
}


void my_Diamondxy( float wx, float wy, float h, int order )
{
    VMvect p[10];
    p[3] = vec_xyz( -wx, 0.0, 0.0 );
    p[2] = vec_xyz( 0.0,  wy, 0.0 );
    p[1] = vec_xyz(  wx, 0.0, 0.0 );
    p[0] = vec_xyz( 0.0, -wy, 0.0 );
    p[4] = vec_xyz( 0.0, 0.0,   h );

    glBegin(GL_TRIANGLES);
       autonormalize_triangle_round(p[0],p[1],p[4],order,-0.002);
       autonormalize_triangle_round(p[1],p[2],p[4],order,-0.002);
       autonormalize_triangle_round(p[2],p[3],p[4],order,-0.002);
       autonormalize_triangle_round(p[3],p[0],p[4],order,-0.002);
    glEnd();
}


void my_Rectxy( float x1, float y1, float z1,    /* inside up */
                float x2, float y2, float z2,
                int fall_x, int order )   /* outside down */
{
    VMvect p[10];

    if(fall_x){
        p[0]=vec_xyz(x1,y1,z1);
        p[1]=vec_xyz(x1,y2,z1);
        p[2]=vec_xyz(x2,y2,z2);
        p[3]=vec_xyz(x2,y1,z2);
        glBegin(GL_QUADS);
          autonormalize_quad(p[0],p[1],p[2],p[3],order);
        glEnd();
    } else {
        p[0]=vec_xyz(x1,y1,z1);
        p[1]=vec_xyz(x1,y2,z2);
        p[2]=vec_xyz(x2,y2,z2);
        p[3]=vec_xyz(x2,y1,z1);
        glBegin(GL_QUADS);
          autonormalize_quad(p[3],p[2],p[1],p[0],!order);
        glEnd();
    }
}


double r1func( double phi )
{
//    return(HOLE1_W/SQR2);
    return(HOLE1_W/SQR2+(1.0-pow((phi-M_PI/4.0)/M_PI*4.0,2))*0.02);
}


double r2func( double phi )
{
    return(HOLE1_W/SQR2+FRAME_D-BANDE_D);
}


void my_Edge( int segnr, double  (*r1)(double), double (*r2)(double), int order )
/* edge shows to pos x,y direction*/
{
    int i;
    double phi1,phi2;
    VMvect n00,n10,n20,n30,n40,n0,n1,n2,n3,n4,n5,n6,n7,n8,n9,v0,v1,v2,v3,v4,v5,v6,v7,v8,v9;
    n00 = vec_xyz(-1,0,0);
    n10 = vec_xyz(0,0,1);
    n20 = vec_unit(vec_xyz(FRAME_DH,0,FRAME_D-BANDE_D-FRAME_PHASE));
    n30 = vec_unit(vec_xyz(FRAME_H-FRAME_PHASE,0,-FRAME_PHASE));
    n40 = vec_unit(vec_xyz(FRAME_H-FRAME_PHASE,0,-FRAME_PHASE));
    for(i=0;i<segnr;i++){
        phi1 = M_PI/2.0*(double)i/(double)segnr;
        phi2 = M_PI/2.0*(double)(i+1)/(double)segnr;
        n0 = vec_rotate(n00,vec_xyz(0,0,phi1));
        v0 = vec_xyz(r1(phi1)*cos(phi1),r1(phi1)*sin(phi1),FRAME_DH-0.006);
        n1 = vec_rotate(n10,vec_xyz(0,0,phi1));
        v1 = vec_xyz(r1(phi1)*cos(phi1),r1(phi1)*sin(phi1),FRAME_DH);
        n2 = vec_rotate(n20,vec_xyz(0,0,phi1));
        v2 = vec_xyz((r2(phi1)-FRAME_PHASE)*cos(phi1),(r2(phi1)-FRAME_PHASE)*sin(phi1),0);
        n3 = vec_rotate(n30,vec_xyz(0,0,phi1));
        v3 = vec_xyz(r2(phi1)*cos(phi1),r2(phi1)*sin(phi1),-FRAME_PHASE);
        n4 = vec_rotate(n40,vec_xyz(0,0,phi1));
        v4 = vec_xyz((r2(phi1)-FRAME_PHASE)*cos(phi1),(r2(phi1)-FRAME_PHASE)*sin(phi1),-FRAME_H);
        n5 = vec_rotate(n10,vec_xyz(0,0,phi2));
        v5 = vec_xyz(r1(phi2)*cos(phi2),r1(phi2)*sin(phi2),FRAME_DH);
        n6 = vec_rotate(n20,vec_xyz(0,0,phi2));
        v6 = vec_xyz((r2(phi2)-FRAME_PHASE)*cos(phi2),(r2(phi2)-FRAME_PHASE)*sin(phi2),0);
        n7 = vec_rotate(n30,vec_xyz(0,0,phi2));
        v7 = vec_xyz(r2(phi2)*cos(phi2),r2(phi2)*sin(phi2),-FRAME_PHASE);
        n8 = vec_rotate(n40,vec_xyz(0,0,phi2));
        v8 = vec_xyz((r2(phi2)-FRAME_PHASE)*cos(phi2),(r2(phi2)-FRAME_PHASE)*sin(phi2),-FRAME_H);
        n9 = vec_rotate(n00,vec_xyz(0,0,phi2));
        v9 = vec_xyz(r1(phi2)*cos(phi2),r1(phi2)*sin(phi2),FRAME_DH-0.006);
        if(!order){
            glBegin(GL_QUAD_STRIP);
/*            glNormal3f(n0.x,n0.y,n0.z);
            glVertex3f(v0.x,v0.y,v0.z);
            glNormal3f(n9.x,n9.y,n9.z);
            glVertex3f(v9.x,v9.y,v9.z);*/
            glNormal3f(n1.x,n1.y,n1.z);
            glVertex3f(v1.x,v1.y,v1.z);
            glNormal3f(n5.x,n5.y,n5.z);
            glVertex3f(v5.x,v5.y,v5.z);
            glNormal3f(n2.x,n2.y,n2.z);
            glVertex3f(v2.x,v2.y,v2.z);
            glNormal3f(n6.x,n6.y,n6.z);
            glVertex3f(v6.x,v6.y,v6.z);
            glNormal3f(n3.x,n3.y,n3.z);
            glVertex3f(v3.x,v3.y,v3.z);
            glNormal3f(n7.x,n7.y,n7.z);
            glVertex3f(v7.x,v7.y,v7.z);
            glNormal3f(n4.x,n4.y,n4.z);
            glVertex3f(v4.x,v4.y,v4.z);
            glNormal3f(n8.x,n8.y,n8.z);
            glVertex3f(v8.x,v8.y,v8.z);
            glEnd();
        } else {
            glBegin(GL_QUAD_STRIP);
/*            glNormal3f(n9.x,n9.y,n9.z);
            glVertex3f(v9.x,v9.y,v9.z);
            glNormal3f(n0.x,n0.y,n0.z);
            glVertex3f(v0.x,v0.y,v0.z);*/
            glNormal3f(n5.x,n5.y,n5.z);
            glVertex3f(v5.x,v5.y,v5.z);
            glNormal3f(n1.x,n1.y,n1.z);
            glVertex3f(v1.x,v1.y,v1.z);
            glNormal3f(n6.x,n6.y,n6.z);
            glVertex3f(v6.x,v6.y,v6.z);
            glNormal3f(n2.x,n2.y,n2.z);
            glVertex3f(v2.x,v2.y,v2.z);
            glNormal3f(n7.x,n7.y,n7.z);
            glVertex3f(v7.x,v7.y,v7.z);
            glNormal3f(n3.x,n3.y,n3.z);
            glVertex3f(v3.x,v3.y,v3.z);
            glNormal3f(n8.x,n8.y,n8.z);
            glVertex3f(v8.x,v8.y,v8.z);
            glNormal3f(n4.x,n4.y,n4.z);
            glVertex3f(v4.x,v4.y,v4.z);
            glEnd();
        }
    }
}


double r1coverfunc( double y )
{
    return( (1.0-y*y)*0.04 );
}


void my_Cover( int segnr, double (*r1)(double), double l, int order )
/* edge shows to pos x,y direction*/
{
    int i;
    double y1,y2;
    VMvect n1,n2,n3,n4,v1,v2,v3,v4,v5,v6,v7,v8;
    double r2;
    r2 = FRAME_D-BANDE_D;
    n1 = vec_xyz(0,0,1);
    n2 = vec_unit(vec_xyz(FRAME_DH,0,FRAME_D-BANDE_D-FRAME_PHASE));
    n3 = vec_unit(vec_xyz(FRAME_H-FRAME_PHASE,0,-FRAME_PHASE));
    n4 = vec_unit(vec_xyz(FRAME_H-FRAME_PHASE,0,-FRAME_PHASE));
    for(i=0;i<segnr;i++){
        y1 = -l/2.0+l*(double)i/(double)segnr;
        y2 = -l/2.0+l*(double)(i+1)/(double)segnr;
        v1 = vec_xyz(r1(y1/l*2.0),y1,FRAME_DH);
        v2 = vec_xyz(r2-FRAME_PHASE,y1,0);
        v3 = vec_xyz(r2,y1,-FRAME_PHASE);
        v4 = vec_xyz(r2-FRAME_PHASE,y1,-FRAME_H);
        v5 = vec_xyz(r1(y2/l*2.0),y2,FRAME_DH);
        v6 = vec_xyz(r2-FRAME_PHASE,y2,0);
        v7 = vec_xyz(r2,y2,-FRAME_PHASE);
        v8 = vec_xyz(r2-FRAME_PHASE,y2,-FRAME_H);
        if(!order){
            glBegin(GL_QUAD_STRIP);
            glNormal3f(n1.x,n1.y,n1.z);
            glVertex3f(v1.x,v1.y,v1.z);
            glNormal3f(n1.x,n1.y,n1.z);
            glVertex3f(v5.x,v5.y,v5.z);
            glNormal3f(n2.x,n2.y,n2.z);
            glVertex3f(v2.x,v2.y,v2.z);
            glNormal3f(n2.x,n2.y,n2.z);
            glVertex3f(v6.x,v6.y,v6.z);
            glNormal3f(n3.x,n3.y,n3.z);
            glVertex3f(v3.x,v3.y,v3.z);
            glNormal3f(n3.x,n3.y,n3.z);
            glVertex3f(v7.x,v7.y,v7.z);
            glNormal3f(n4.x,n4.y,n4.z);
            glVertex3f(v4.x,v4.y,v4.z);
            glNormal3f(n4.x,n4.y,n4.z);
            glVertex3f(v8.x,v8.y,v8.z);
            glEnd();
        } else {
            glBegin(GL_QUAD_STRIP);
            glNormal3f(n1.x,n1.y,n1.z);
            glVertex3f(v5.x,v5.y,v5.z);
            glNormal3f(n1.x,n1.y,n1.z);
            glVertex3f(v1.x,v1.y,v1.z);
            glNormal3f(n2.x,n2.y,n2.z);
            glVertex3f(v6.x,v6.y,v6.z);
            glNormal3f(n2.x,n2.y,n2.z);
            glVertex3f(v2.x,v2.y,v2.z);
            glNormal3f(n3.x,n3.y,n3.z);
            glVertex3f(v7.x,v7.y,v7.z);
            glNormal3f(n3.x,n3.y,n3.z);
            glVertex3f(v3.x,v3.y,v3.z);
            glNormal3f(n4.x,n4.y,n4.z);
            glVertex3f(v8.x,v8.y,v8.z);
            glNormal3f(n4.x,n4.y,n4.z);
            glVertex3f(v4.x,v4.y,v4.z);
            glEnd();
        }
    }
}


void my_Bande( float x1, float y1, float z1,    /* inside up */
               float x2, float y2, float z2,    /* outside down */
               float tan1, float tan2, int order )
{/*   ^ y                        *
  *   |                          *
  *   .z -> x                    *
  *                              *
  *   4  8               9  7    *
  * 0 \                     / 3  *
  *     \5_______________6/      *
  *     1                 2      */
    float dummy;
    VMvect p[10];

    p[0]=vec_xyz(x1-(y2-y1)*tan1,y2,z1+FRAME_DH);
    p[1]=vec_xyz(x1,y1,z1-(z1-z2)*0.1);
    p[2]=vec_xyz(x2,y1,z1-(z1-z2)*0.1);
    p[3]=vec_xyz(x2+(y2-y1)*tan2,y2,z1+FRAME_DH);
    p[4]=vec_xyz(x1-(y2-y1)*tan1,y2,z2);
    p[5]=vec_xyz(x1,y1+(y2-y1)*0.3,z2);
    p[6]=vec_xyz(x2,y1+(y2-y1)*0.3,z2);
    p[7]=vec_xyz(x2+(y2-y1)*tan2,y2,z2);
    p[8]=vec_xyz(x1,y2,z1+FRAME_DH);
    p[9]=vec_xyz(x2,y2,z1+FRAME_DH);

    if( x1 > x2 ){ dummy=x1; x1=x2; x2=dummy; }
    if( y1 > y2 ){ dummy=y1; y1=y2; y2=dummy; }
    if( z1 > z2 ){ dummy=z1; z1=z2; z2=dummy; }
    
    glBegin(GL_QUADS);
       autonormalize_quad(p[1],p[2],p[9],p[8],order);
       autonormalize_quad(p[5],p[6],p[2],p[1],order);
    glEnd();

    glBegin(GL_TRIANGLES);
       autonormalize_triangle(p[0],p[1],p[8],order);
       autonormalize_triangle(p[0],p[5],p[1],order);
       autonormalize_triangle(p[0],p[4],p[5],order);
       autonormalize_triangle(p[3],p[9],p[2],order);
       autonormalize_triangle(p[3],p[2],p[6],order);
       autonormalize_triangle(p[3],p[6],p[7],order);
    glEnd();
}


int create_table( int reflect_bind, BordersType *borders )
{
    int table_obj, i;
    int frametexbind,frametexw,frametexh;
    int tabletexbind,tabletexw,tabletexh;
    int depth;
    char * frametexdata;
    char * tabletexdata;
    float balld  = BALL_D/**100.0*/;   // cm
    float tablew = TABLE_W/**100.0*/;  // cm
    float tablel = TABLE_L/**100.0*/;  // cm
    float tableh = TABLE_H/**100.0*/;  // cm
    float bande_d = BANDE_D; // m
    float cm  =  0.01;  // cm
    double hole_r1,hole_r2;
    GLfloat tab_col_spec[4] = {0.0, 0.0, 0.0, 0.0};
    GLfloat tab_col_diff[4] = {0.05, 0.4, 0.13, 1.0};
    GLfloat tab_col_amb [4];
    GLfloat wood_col_spec[4] = {1.0, 1.0, 1.0, 1.0};
    GLfloat wood_col_diff[4] = {0.25, 0.08, 0.02, 1.0};
    GLfloat wood_col_diff2[4] = {0.8, 0.8, 0.8, 1.0};
    GLfloat wood_col_amb [4];
    GLfloat wood_col_shin = 10.0;
    GLfloat dia_col_spec[4] = {1.0, 0.9, 0.4, 1.0};
    GLfloat dia_col_diff[4] = {0.8, 0.7, 0.1, 1.0};
    GLfloat dia_col_amb [4];
    GLfloat dia_col_shin = 0.0;
    GLfloat t_gen_params[] = {8.0,0.0,0.0,0.0};
    GLfloat s_gen_params[] = {0.0,1.0,0.0,0.0};
    GLfloat s_gen_params2[] = {1.0,0.0,0.0,0.0};
    GLfloat t_gen_params2[] = {0.0,8.0,0.0,0.0};

    if( options_frame_tex ) {
        glGenTextures(1,&frametexbind);
        load_png("table-frame.png",&frametexw,&frametexh,&depth,&frametexdata);
        glBindTexture(GL_TEXTURE_2D,frametexbind);
        gluBuild2DMipmaps(GL_TEXTURE_2D, 3, frametexw, frametexh, GL_RGB,
                          GL_UNSIGNED_BYTE, frametexdata);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, options_tex_min_filter);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, options_tex_mag_filter);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
        free( frametexdata );
    } else {
        frametexbind=0;
    }

    if( options_table_tex ) {
        glGenTextures(1,&tabletexbind);
        load_png("tabletex_fB_256x256.png",&tabletexw,&tabletexh,&depth,&tabletexdata);
//        load_png("table-frame.png",&tabletexw,&tabletexh,&depth,&tabletexdata);
        glBindTexture(GL_TEXTURE_2D,tabletexbind);
        gluBuild2DMipmaps(GL_TEXTURE_2D, 1, tabletexw, tabletexh, GL_LUMINANCE,
                          GL_UNSIGNED_BYTE, tabletexdata);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, options_tex_min_filter);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, options_tex_mag_filter);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
        free( tabletexdata );
    } else {
        tabletexbind=0;
    }

   for(i=0;i<3;i++) tab_col_amb[i]=tab_col_diff[i]*0.5;
   for(i=0;i<3;i++) wood_col_amb[i]=wood_col_diff[i]*0.5;
   for(i=0;i<3;i++) dia_col_amb[i]=dia_col_diff[i]*0.5;

   table_obj = glGenLists(1);
   glNewList(table_obj, GL_COMPILE);
//   glScalef( 1.0, 1.0, 1.0 );

   glMaterialfv(GL_FRONT, GL_DIFFUSE,   tab_col_diff);
   glMaterialfv(GL_FRONT, GL_AMBIENT,   tab_col_amb);
   glMaterialfv(GL_FRONT, GL_SPECULAR,  tab_col_spec);
   glMaterialf (GL_FRONT, GL_SHININESS, 0.0 );

   //glTranslatef( 0, 0, 0 );

   //gluQuadricDrawStyle(q, GLU_LINE);
   /* cylinder */
//   gluQuadricNormals(q, GL_SMOOTH);
//   gluQuadricTexture(q, GL_TRUE);
   //gluCylinder(q, 0.5, 0.5, 2.0, 24, 1);
   
   //glTranslatef(1.0, 0.0, 0.5);
   //gluCylinder(q, 0.0, 0.4, 1.0, 24, 1);
   //glTranslatef(-1.0, 0.0, -0.5);

   /*glTranslatef( 0, 0, balld/2.0 );
   glRectd( -tablew/2.0, -tablel/2.0, tablew/2.0, tablel/2.0 );
   glTranslatef( 0, 0, tableh );
   glRectd( tablew/2.0, -tablel/2.0, -tablew/2.0, tablel/2.0 );
   glTranslatef( 0, 0, -tableh-balld/2.0 );*/

   if( options_table_tex ){
       glEnable(GL_TEXTURE_2D);
       glBindTexture(GL_TEXTURE_2D,tabletexbind);
   } else {
       glDisable(GL_TEXTURE_2D);
   }
/*   glEnable(GL_TEXTURE_GEN_S);
   glEnable(GL_TEXTURE_GEN_T);
   glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
   glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
   glTexGenfv(GL_S, GL_OBJECT_PLANE, s_gen_params);
   glTexGenfv(GL_T, GL_OBJECT_PLANE, t_gen_params);*/
   glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

//   glShadeModel(GL_FLAT);
   glBegin(GL_QUADS);
      glTexCoord2f( -0.7-0.2, -0.2 );
      glNormal3f( 0.0, 0.0, 1.0 );
      glVertex3f(-tablew/2.0, -tablel/2.0, -balld/2.0);
      glTexCoord2f( 1.7-0.2, -0.2 );
      glNormal3f( 0.0, 0.0, 1.0 );
      glVertex3f(-tablew/2.0, +tablel/2.0, -balld/2.0);
      glTexCoord2f( 1.7+0.2, 1.2 );
      glNormal3f( 0.0, 0.0, 1.0 );
      glVertex3f(+tablew/2.0, +tablel/2.0, -balld/2.0);
      glTexCoord2f( -0.7+0.2, 1.2 );
      glNormal3f( 0.0, 0.0, 1.0 );
      glVertex3f(+tablew/2.0, -tablel/2.0, -balld/2.0);
   glEnd();
/*   my_glBox( -tablew/2.0, -tablel/2.0, -balld/2.0,
              tablew/2.0,  tablel/2.0, -balld/2.0-6.0*cm );*/

   hole_r1 = HOLE1_W/SQR2 /**100.0*/;
   hole_r2 = HOLE2_W/2.0  /**100.0*/;

/*   my_glBox( -tablew/2.0-5.0*cm, -hole_r2, -balld/2.0-5.0*cm,
             -tablew/2.0,    -tablel/2.0+hole_r1, 0.0 );

   my_glBox( -tablew/2.0-5.0*cm, +hole_r2, -balld/2.0-5.0*cm,
             -tablew/2.0,   +tablel/2.0-hole_r1, 0.0 );

   my_glBox( +tablew/2.0+5.0*cm, -hole_r2, -balld/2.0-5.0*cm,
             +tablew/2.0,   -tablel/2.0+hole_r1, 0.0 );

   my_glBox( +tablew/2.0+5.0*cm, +hole_r2, -balld/2.0-5.0*cm,
             +tablew/2.0,   +tablel/2.0-hole_r1, 0.0 );

   my_glBox( -tablew/2.0+hole_r1, -tablel/2.0-5.0*cm, -balld/2.0-5.0*cm,
             +tablew/2.0-hole_r1, -tablel/2.0,   0.0 );

   my_glBox( -tablew/2.0+hole_r1, +tablel/2.0+5.0*cm, -balld/2.0-5.0*cm,
             +tablew/2.0-hole_r1, +tablel/2.0,   0.0 );
*/

   /* upper */
   my_Bande( -tablew/2.0+hole_r1, tablel/2.0,        0.0,
             +tablew/2.0-hole_r1, tablel/2.0+bande_d, -balld/2.0,
             HOLE1_TAN, HOLE1_TAN, 0 );

   /* lower */
   glPushMatrix();
   glScalef(1.0,-1.0,1.0);
   my_Bande( -tablew/2.0+hole_r1, tablel/2.0,        0.0,
             +tablew/2.0-hole_r1, tablel/2.0+bande_d, -balld/2.0,
             HOLE1_TAN, HOLE1_TAN, 1 );
   glPopMatrix();

   glPushMatrix();
   glRotatef( 90.0, 0.0,0.0,1.0 );
   /* upper left */
   my_Bande( hole_r2,         tablew/2.0,        0.0,
             +tablew-hole_r1, tablew/2.0+bande_d, -balld/2.0,
             HOLE2_TAN, HOLE1_TAN, 0 );
   glScalef(1.0,-1.0,1.0);
   /* upper right */
   my_Bande( hole_r2,         tablew/2.0,        0.0,
             +tablew-hole_r1, tablew/2.0+bande_d, -balld/2.0,
             HOLE2_TAN, HOLE1_TAN, 1 );
   glPopMatrix();

   glPushMatrix();
   glRotatef( 90.0, 0.0,0.0,1.0 );
   /* lower left */
   my_Bande( -tablew+hole_r1,         tablew/2.0,        0.0,
             -hole_r2, tablew/2.0+bande_d, -balld/2.0,
             HOLE1_TAN, HOLE2_TAN, 0 );
   glScalef(1.0,-1.0,1.0);
   /* lower right */
   my_Bande( -tablew+hole_r1,         tablew/2.0,        0.0,
             -hole_r2, tablew/2.0+bande_d, -balld/2.0,
             HOLE1_TAN, HOLE2_TAN, 1 );
   glPopMatrix();

   /* diamonds */
   glMaterialfv(GL_FRONT, GL_DIFFUSE,   dia_col_diff);
   glMaterialfv(GL_FRONT, GL_AMBIENT,   dia_col_amb);
   glMaterialfv(GL_FRONT, GL_SPECULAR,  dia_col_spec);
   glMaterialf (GL_FRONT, GL_SHININESS, dia_col_shin);
   if( options_diamond_reflect ){
       glEnable(GL_TEXTURE_2D);
       glBindTexture(GL_TEXTURE_2D,reflect_bind);
       glEnable(GL_TEXTURE_GEN_S);
       glEnable(GL_TEXTURE_GEN_T);
       glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
       glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
       glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
   } else {
       glDisable(GL_TEXTURE_2D);
   }

   /*left right*/
   for(i=0;i<4;i++){
       int flip=0;
       glPushMatrix();
       switch(i){
       case 0: flip=0; break;
       case 1: glScalef( -1.0,  1.0, 1.0 ); flip=1; break;
       case 2: glScalef(  1.0, -1.0, 1.0 ); flip=1; break;
       case 3: glScalef( -1.0, -1.0, 1.0 ); flip=0; break;
       }
       glPushMatrix();
       glTranslatef( tablew/2.0+FRAME_D/2.0-FRAME_PHASE/2.0+bande_d/2.0, 2.0*tablel/8.0, FRAME_DH/2.0 );
       glRotatef( atan(FRAME_DH/(FRAME_D-bande_d-FRAME_PHASE))*180.0/M_PI, 0.0, 1.0, 0.0 );
       my_Diamondxy( 0.02, 0.014, 0.006, flip );
       glPopMatrix();
       glPushMatrix();
       glTranslatef( tablew/2.0+FRAME_D/2.0-FRAME_PHASE/2.0+bande_d/2.0, 3.0*tablel/8.0, FRAME_DH/2.0 );
       glRotatef( atan(FRAME_DH/(FRAME_D-bande_d-FRAME_PHASE))*180.0/M_PI, 0.0, 1.0, 0.0 );
       my_Diamondxy( 0.02, 0.014, 0.006, flip );
       glPopMatrix();
       glPushMatrix();
       glTranslatef( tablew/2.0+FRAME_D/2.0-FRAME_PHASE/2.0+bande_d/2.0, 1.0*tablel/8.0, FRAME_DH/2.0 );
       glRotatef( atan(FRAME_DH/(FRAME_D-bande_d-FRAME_PHASE))*180.0/M_PI, 0.0, 1.0, 0.0 );
       my_Diamondxy( 0.02, 0.014, 0.006, flip );
       glPopMatrix();
       glPopMatrix();
   }
   /*upper lower*/
   for(i=0;i<2;i++){
       int flip=0;
       glPushMatrix();
       switch(i){
       case 0: flip=0; break;
       case 1: glScalef( 1.0, -1.0, 1.0 ); flip=1; break;
       }
       glPushMatrix();
       glTranslatef( -tablew/2.0+1.0*tablew/4.0, tablel/2.0+FRAME_D/2.0-FRAME_PHASE/2.0+bande_d/2.0, FRAME_DH/2.0 );
       glRotatef( -atan(FRAME_DH/(FRAME_D-bande_d-FRAME_PHASE))*180.0/M_PI, 1.0, 0.0, 0.0 );
       my_Diamondxy( 0.014, 0.02, 0.006, flip );
       glPopMatrix();
       glPushMatrix();
       glTranslatef( -tablew/2.0+2.0*tablew/4.0, tablel/2.0+FRAME_D/2.0-FRAME_PHASE/2.0+bande_d/2.0, FRAME_DH/2.0 );
       glRotatef( -atan(FRAME_DH/(FRAME_D-bande_d-FRAME_PHASE))*180.0/M_PI, 1.0, 0.0, 0.0 );
       my_Diamondxy( 0.014, 0.02, 0.006, flip );
       glPopMatrix();
       glPushMatrix();
       glTranslatef( -tablew/2.0+3.0*tablew/4.0, tablel/2.0+FRAME_D/2.0-FRAME_PHASE/2.0+bande_d/2.0, FRAME_DH/2.0 );
       glRotatef( -atan(FRAME_DH/(FRAME_D-bande_d-FRAME_PHASE))*180.0/M_PI, 1.0, 0.0, 0.0 );
       my_Diamondxy( 0.014, 0.02, 0.006, flip );
       glPopMatrix();
       glPopMatrix();
   }

//   glShadeModel(GL_SMOOTH);
   /* edges */
   glPushMatrix();
     glTranslatef( tablew/2.0-HOLE1_W/SQR2+bande_d*HOLE1_TAN, tablel/2.0-HOLE1_W/SQR2+bande_d*HOLE1_TAN, 0.0 );
     my_Edge( 6, r1func, r2func, 0 );
   glPopMatrix();
   glPushMatrix();
     glScalef( -1.0, +1.0, +1.0 );
     glTranslatef( tablew/2.0-HOLE1_W/SQR2+bande_d*HOLE1_TAN, tablel/2.0-HOLE1_W/SQR2+bande_d*HOLE1_TAN, 0.0 );
     my_Edge( 6, r1func, r2func, 1 );
   glPopMatrix();
   glPushMatrix();
     glScalef( -1.0, -1.0, +1.0 );
     glTranslatef( tablew/2.0-HOLE1_W/SQR2+bande_d*HOLE1_TAN, tablel/2.0-HOLE1_W/SQR2+bande_d*HOLE1_TAN, 0.0 );
     my_Edge( 6, r1func, r2func, 0 );
   glPopMatrix();
   glPushMatrix();
     glScalef( +1.0, -1.0, +1.0 );
     glTranslatef( tablew/2.0-HOLE1_W/SQR2+bande_d*HOLE1_TAN, tablel/2.0-HOLE1_W/SQR2+bande_d*HOLE1_TAN, 0.0 );
     my_Edge( 6, r1func, r2func, 1 );
   glPopMatrix();
   /* covers */
   glPushMatrix();
     glTranslatef( tablew/2.0+bande_d, 0.0, 0.0 );
     my_Cover( 4, r1coverfunc, HOLE2_W-2.0*BANDE_D*HOLE2_TAN, 0 );
   glPopMatrix();
   glPushMatrix();
     glScalef( -1.0, +1.0, +1.0 );
     glTranslatef( tablew/2.0+bande_d, 0.0, 0.0 );
     my_Cover( 4, r1coverfunc, HOLE2_W-2.0*BANDE_D*HOLE2_TAN, 1 );
   glPopMatrix();

//   glShadeModel(GL_FLAT);
   /* wood-frame */
   glMaterialfv(GL_FRONT, GL_AMBIENT,   wood_col_amb);
   glMaterialfv(GL_FRONT, GL_SPECULAR,  wood_col_spec);
   glMaterialf (GL_FRONT, GL_SHININESS, wood_col_shin);

   if( options_frame_tex ){
       glEnable(GL_TEXTURE_2D);
       glMaterialfv(GL_FRONT, GL_DIFFUSE,   wood_col_diff2);
       glBindTexture(GL_TEXTURE_2D,frametexbind);
       glEnable(GL_TEXTURE_GEN_S);
       glEnable(GL_TEXTURE_GEN_T);
       glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
       glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
       glTexGenfv(GL_S, GL_OBJECT_PLANE, s_gen_params);
       glTexGenfv(GL_T, GL_OBJECT_PLANE, t_gen_params);
       glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
   } else {
       glMaterialfv(GL_FRONT, GL_DIFFUSE,   wood_col_diff);
       glDisable(GL_TEXTURE_2D);
   }
   /*left right*/
   for(i=0;i<4;i++){
       int flip=0;
       glPushMatrix();
       switch(i){
       case 0: flip=0; break; /* right up */
       case 1: glScalef( -1.0,  1.0, 1.0 ); flip=1; break;
       case 2: glScalef(  1.0, -1.0, 1.0 ); flip=1; break;
       case 3: glScalef( -1.0, -1.0, 1.0 ); flip=0; break;
       }
       my_Rectxy( tablew/2.0+bande_d, tablel/2.0-hole_r1+bande_d*HOLE1_TAN, FRAME_DH,
                  tablew/2.0+FRAME_D-FRAME_PHASE, HOLE2_W/2.0-bande_d*HOLE2_TAN,        0.0,
                  1, flip );
       my_Rectxy( tablew/2.0+FRAME_D-FRAME_PHASE, tablel/2.0-hole_r1+bande_d*HOLE1_TAN, 0.0,
                  tablew/2.0+FRAME_D, HOLE2_W/2.0-bande_d*HOLE2_TAN,        -FRAME_PHASE,
                  1, flip );
       my_Rectxy( tablew/2.0+FRAME_D, tablel/2.0-hole_r1+bande_d*HOLE1_TAN, -FRAME_PHASE,
                  tablew/2.0+FRAME_D-FRAME_PHASE, HOLE2_W/2.0-bande_d*HOLE2_TAN, -FRAME_H,
                  1, flip );
       glPopMatrix();
   }
   /*upper lower*/
   glTexGenfv(GL_S, GL_OBJECT_PLANE, s_gen_params2);
   glTexGenfv(GL_T, GL_OBJECT_PLANE, t_gen_params2);
   for(i=0;i<2;i++){
       int flip=0;
       glPushMatrix();
       switch(i){
       case 0: flip=1; break;
       case 1: glScalef( 1.0, -1.0, 1.0 ); flip=0; break;
       }
       my_Rectxy( -tablew/2.0+hole_r1-bande_d*HOLE1_TAN, tablel/2.0+bande_d, FRAME_DH,
                  +tablew/2.0-hole_r1+bande_d*HOLE1_TAN, tablel/2.0+FRAME_D-FRAME_PHASE, 0.0,
                  0, flip );
       my_Rectxy( -tablew/2.0+hole_r1-bande_d*HOLE1_TAN, tablel/2.0+FRAME_D-FRAME_PHASE, 0.0,
                  +tablew/2.0-hole_r1+bande_d*HOLE1_TAN, tablel/2.0+FRAME_D, -FRAME_PHASE,
                  0, flip );
       my_Rectxy( -tablew/2.0+hole_r1-bande_d*HOLE1_TAN, tablel/2.0+FRAME_D, -FRAME_PHASE,
                  +tablew/2.0-hole_r1+bande_d*HOLE1_TAN, tablel/2.0+FRAME_D-FRAME_PHASE, -FRAME_H,
                  0, flip );
       glPopMatrix();
   }

   glDisable(GL_TEXTURE_2D);
   glMaterialfv(GL_FRONT, GL_DIFFUSE,   wood_col_diff);
   my_glBox( -tablew/2.0*0.8-3.0*cm, -tablel/2.0*0.8-3.0*cm, -balld/2.0-6.0*cm,
             +tablew/2.0*0.8+3.0*cm, +tablel/2.0*0.8+3.0*cm, -tableh-balld/2.0 );

//   glEnable(GL_TEXTURE_2D);

/*   for(i=0;i<borders->holenr;i++){
       GLUquadric * qd;
       qd=gluNewQuadric();
       glPushMatrix();
       glTranslatef( borders->hole[i].pos.x, borders->hole[i].pos.y, borders->hole[i].pos.z );
       gluSphere( qd, borders->hole[i].r, 8, 8 );
       glPopMatrix();
   }*/

//   glShadeModel(GL_SMOOTH);

   glEndList();

   return table_obj;

}
